#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>
#include <GL/glx.h>
#include <sys/time.h>

#include "engine.h"
#include "debug.h"

Bool XkbSetDetectableAutoRepeat(Display *p_display, Bool detectable, Bool *p_supported);

Display *engine_p_display = NULL;
Screen *engine_p_screen;
int engine_screen_number;
Window engine_window = 0;

unsigned
	engine_width = 640,
	engine_height = 480;
char
	engine_depth = 16,
	engine_fullscreen_flag = 0;
int engine_real_depth = 0;

char *engine_title = NULL;

unsigned long *engine_icon_data = NULL;
int engine_icon_data_length = 0;

GLXContext engine_glx_context = NULL;
GLint engine_gl_max_size[2];

void engine_default_drawing_function() {}
void (*engine_drawing_function)() = engine_default_drawing_function;
void (*engine_keys_handler)(unsigned keycode, int x, int y, char state_flag) = NULL;
void (*engine_motion_handler)(int x, int y) = NULL;
void (*engine_focus_handler)(char state) = NULL;

struct timeval engine_min_delta_tv = {.tv_sec = 0, .tv_usec = 16666}; /* 60  fps */

enum {engine_wm_delete_window, engine_wm_save_yourself};
Atom
	engine_atoms[2] = {None, None},
	engine_atom_wm_protocols = None,
	engine_atom_wm_normal_hints = None;

/* FIXME add graphical error message */
void engine_error_exit(char *message)
{
	if (message != NULL)
		fprintf(stderr, "Error: %s\n", message);
	else
		perror("Error");
	exit(EXIT_FAILURE);
}

void engine_calc_pos_to_be_center(int *x, int *y, unsigned width, unsigned height)
{
	*x = (WidthOfScreen(engine_p_screen) - width)/2;
	*y = (HeightOfScreen(engine_p_screen) - height)/2;
	if (*x < 0) *x = 0;
	if (*y < 0) *y = 0;
}

int engine_set_size_hints()
{
	XSizeHints *p_size_hints = XAllocSizeHints();
	if (p_size_hints == NULL)
		return engine_ec_error;
	p_size_hints->flags = PPosition | PMinSize | PMaxSize;
	p_size_hints->min_width = engine_width;
	p_size_hints->min_height = engine_height;
	p_size_hints->max_width = engine_width;
	p_size_hints->max_height = engine_height;
	if (engine_atom_wm_normal_hints == None)
		engine_atom_wm_normal_hints = XInternAtom
		(
			engine_p_display,
			"WM_NORMAL_HINTS",
			False
		);
	XSetWMSizeHints
	(
		engine_p_display,
		engine_window,
		p_size_hints,
		engine_atom_wm_normal_hints
	);
	XFree(p_size_hints);
	return engine_ec_ok;
}

void engine_apply_title()
{
	if (engine_window != 0 && engine_title != NULL)
	{
		Atom atom__net_wm_name = XInternAtom(engine_p_display, "_NET_WM_NAME", False);
		Atom atom_utf8_string = XInternAtom(engine_p_display, "UTF8_STRING", False);
		XChangeProperty
		(
			engine_p_display,
			engine_window,
			atom__net_wm_name,
			atom_utf8_string,
			8,
			PropModeReplace,
			(unsigned char *)engine_title,
			strlen(engine_title)
		);
	}
}

void engine_apply_icon()
{
	if
	(
		engine_window != 0 &&
		engine_icon_data != NULL &&
		engine_icon_data_length != 0
	)
	{
		Atom atom__net_wm_icon = XInternAtom(engine_p_display, "_NET_WM_ICON", False);
		XChangeProperty
		(
			engine_p_display,
			engine_window,
			atom__net_wm_icon,
			XA_CARDINAL,
			32,
			PropModeReplace,
			(unsigned char *)engine_icon_data,
			engine_icon_data_length
		);
	}
}

long engine_calc_event_mask()
{
	return (engine_keys_handler != NULL ?
		(
			KeyPressMask | KeyReleaseMask |
			ButtonPressMask | ButtonReleaseMask |
			FocusChangeMask
		) : 0) |
		(engine_motion_handler != NULL ? PointerMotionMask : 0) |
		(engine_focus_handler != NULL ?
			(FocusChangeMask | EnterWindowMask | LeaveWindowMask) : 0);
}

void engine_set_adequate_event_mask()
{
	if (engine_window != 0)
	{
		XSetWindowAttributes window_attributes =
			{ .event_mask = engine_calc_event_mask() };
		XChangeWindowAttributes(engine_p_display, engine_window, CWEventMask, &window_attributes);
	}
}

int engine_check_that_gl_engine_support_size(unsigned width, unsigned height)
{
	if (width > engine_gl_max_size[0] || height > engine_gl_max_size[1])
		return engine_ec_error;
	return engine_ec_ok;
}

void engine_glx_choose_visual(XVisualInfo **p_visual_info)
{
	int attribute_list[] =
	{
		GLX_RGBA,
		GLX_DOUBLEBUFFER,
		GLX_BUFFER_SIZE, engine_depth,
		None
	};
	*p_visual_info = glXChooseVisual
	(
		engine_p_display,
		engine_screen_number,
		attribute_list
	);
}

int engine_create_window()
{
	if (engine_p_display == NULL)
		return engine_ec_error;
	if (engine_window != 0)
		return engine_ec_window_already_created;

	XVisualInfo *p_visual_info;
	engine_glx_choose_visual(&p_visual_info);
	if (p_visual_info == NULL)
		return engine_ec_bad_depth;
	engine_real_depth = p_visual_info->depth;

	Window root_window = XRootWindow(engine_p_display, engine_screen_number);
	XSetWindowAttributes window_attributes =
	{
		.background_pixel = 0,
		.border_pixel = 0,
		.colormap = XCreateColormap
		(
			engine_p_display,
			root_window,
			p_visual_info->visual,
			AllocNone
		),
		.event_mask = engine_calc_event_mask()
	};

	int x, y;
	/* FIXME use also window border size */
	engine_calc_pos_to_be_center(&x, &y, engine_width, engine_height);
	engine_window = XCreateWindow
	(
		engine_p_display,
		root_window,
		x,
		y,
		engine_width,
		engine_height,
		0,
		engine_real_depth,
		InputOutput,
		p_visual_info->visual,
		CWBackPixel | CWBorderPixel | CWColormap | CWEventMask,
		&window_attributes
	);

	if (engine_set_size_hints(engine_width, engine_height) != engine_ec_ok)
		engine_error_exit("Cannot create size hints.");

	engine_apply_title();
	engine_apply_icon();

	if (engine_atoms[engine_wm_delete_window] == None)
		engine_atoms[engine_wm_delete_window] = XInternAtom(engine_p_display, "WM_DELETE_WINDOW", True);
	if (engine_atoms[engine_wm_save_yourself] == None)
		engine_atoms[engine_wm_save_yourself] = XInternAtom(engine_p_display, "WM_SAVE_YOURSELF", True);
	if (engine_atom_wm_protocols == None)
		engine_atom_wm_protocols = XInternAtom (engine_p_display, "WM_PROTOCOLS", True);
	if (engine_atom_wm_protocols != None && engine_atoms[engine_wm_delete_window] != None)
	{
		if (engine_atoms[engine_wm_save_yourself] == None)
			XSetWMProtocols(engine_p_display, engine_window, engine_atoms, 1);
		else
			XSetWMProtocols(engine_p_display, engine_window, engine_atoms, 2);
	}

	if (engine_fullscreen_flag)
	{
		DBGPRINT(FIXME, "use engine_fullscreen_flag");
	}

	engine_glx_context = glXCreateContext(engine_p_display, p_visual_info, NULL, True);
	if (engine_glx_context == NULL)
		engine_error_exit("glXCreateContext failed");
	XFree(p_visual_info);

	if (glXMakeCurrent(engine_p_display, engine_window, engine_glx_context) != True)
		engine_error_exit("Error: glXMakeCurrent failed");
	glGetIntegerv(GL_MAX_VIEWPORT_DIMS, engine_gl_max_size);

	if (engine_check_that_gl_engine_support_size(engine_width, engine_height) != engine_ec_ok)
		return engine_ec_window_size_too_big;

	XMapWindow(engine_p_display, engine_window);
	XPending(engine_p_display); /* to display window */

	return engine_ec_ok;
}

int engine_init()
{
	if (engine_p_display == NULL)
	{
		engine_p_display = XOpenDisplay(NULL);
		if (engine_p_display == NULL)
			engine_error_exit("Can not open display.");

		/* discard dummy KeyRelease before every KeyPress */
		Bool supported;
		if
		(
			XkbSetDetectableAutoRepeat(engine_p_display, True, &supported) != True
			|| supported != True
		)
			engine_error_exit("DetectableAutorepeat not supported");

		engine_p_screen = XDefaultScreenOfDisplay(engine_p_display);
		engine_screen_number = DefaultScreen(engine_p_display);
	}

	int result;
	result = engine_create_window();
	if (result != engine_ec_ok)
		return result;

	Atom
		atom__net_workarea = XInternAtom(engine_p_display, "_NET_WM_STRUT", False),
		actual_type_return;
	int actual_format_return;
	unsigned long
		nitems_return,
		bytes_after_return;
	unsigned char *prop_return;
	XGetWindowProperty
	(
		engine_p_display,
		engine_window,
		atom__net_workarea,
		0,
		16,
		False,
		AnyPropertyType,
		&actual_type_return,
		&actual_format_return,
		&nitems_return,
		&bytes_after_return,
		&prop_return
	);
	DBGPRINTF
	(
		TMP,
		"actual_type_return=%i, actual_format_return=%i, nitems_return=%lu, bytes_after_return=%lu, prop_return=%u",
		actual_type_return,
		actual_format_return,
		nitems_return,
		bytes_after_return,
		prop_return
	);

	return engine_ec_ok;
}

/* FIXME add support for (0,0,0,0) - fullscreen with native resolution@depth */
int engine_set_mode(unsigned width, unsigned height, char depth, char fullscreen_flag)
{
	if (engine_p_display != NULL && engine_window != 0)
	{
		if (fullscreen_flag)
		{
			DBGPRINT(FIXME, "use engine_fullscreen_flag");
		}
		if (depth != engine_depth && depth != engine_real_depth)
		{
			XVisualInfo *p_visual_info;
			engine_glx_choose_visual(&p_visual_info);
			if (depth != p_visual_info->depth)
			{
				XFree(p_visual_info);
				glXMakeCurrent(engine_p_display, None, NULL);
				glXDestroyContext(engine_p_display, engine_glx_context);
				XDestroyWindow(engine_p_display, engine_window);
				engine_window = 0;
				engine_width = width;
				engine_height = height;
				engine_depth = depth;
				int result;
				result = engine_create_window();
				if (result != engine_ec_ok)
					return result;
				return engine_ec_ok;
			}
			XFree(p_visual_info);
		}
		engine_depth = depth;
		if (engine_width != width || engine_height != height)
		{
			if
			(
				width <= engine_width && height <= engine_height ||
				engine_check_that_gl_engine_support_size(width, height) == engine_ec_ok
			)
			{
				int x, y;
				engine_calc_pos_to_be_center(&x, &y, width, height);
				XMoveWindow
				(
					engine_p_display,
					engine_window,
					x,
					y
				);
			}
			else
				return engine_ec_window_size_too_big;
			engine_width = width;
			engine_height = height;
		}
		if (engine_set_size_hints(engine_width, engine_height) != engine_ec_ok)
			engine_error_exit("Cannot create size hints.");
		glViewport(0, 0, engine_width, engine_height);

		XMapWindow(engine_p_display, engine_window);
		XPending(engine_p_display); /* to display window */
		return engine_ec_ok;
	}
	engine_width = width;
	engine_height = height;
	engine_depth = depth;
	engine_fullscreen_flag = fullscreen_flag;
	return engine_ec_ok;
}

int engine_set_title(char *title)
{
	unsigned len = strlen(title) + 1;
	if (engine_title != NULL)
	{
		if (strlen(engine_title) > len)
			goto enough_size;
		free(engine_title);
	}
	engine_title = (char *)malloc(len);
	if (engine_title == NULL)
		return engine_ec_error;
enough_size:
	memcpy(engine_title, title, len);
	engine_apply_title();
	return engine_ec_ok;
}

void engine_set_icon(unsigned long *icon_data, int icon_data_length)
{
	engine_icon_data = icon_data;
	engine_icon_data_length = icon_data_length;
	engine_apply_icon();
}

void engine_set_drawing_function(void (*drawing_function)())
{
	ASSERT(drawing_function != NULL);
	engine_drawing_function = drawing_function;
	engine_set_adequate_event_mask();
}

void engine_set_keys_handler
	(void (*keys_handler)(unsigned keycode, int x, int y, char state_flag))
{
	ASSERT(keys_handler != NULL);
	engine_keys_handler = keys_handler;
	engine_set_adequate_event_mask();
}

void engine_set_motion_handler(void (*motion_handler)(int x, int y))
{
	ASSERT(motion_handler != NULL);
	engine_motion_handler = motion_handler;
	engine_set_adequate_event_mask();
}

void engine_set_focus_handler(void (*focus_handler)(char state))
{
	ASSERT(focus_handler != NULL);
	engine_focus_handler = focus_handler;
	engine_set_adequate_event_mask();
}

void engine_set_max_fps(float max_fps)
{
	max_fps = 1 / max_fps;
	engine_min_delta_tv.tv_sec = max_fps;
	engine_min_delta_tv.tv_usec = 1000000 * (max_fps - (long)max_fps);
}

void engine_main_loop()
{
	XEvent event;
	unsigned keycode;
	char
		single_key_up_flag,
		event_while_waiting_flag;
	struct timeval tv, delta_tv;
	int
		xserver_fd,
		xserver_fd_p1;
	fd_set fd_set_read;

	memset(engine_keys_states, 1, engine_keys_names_number);
	gettimeofday(&tv, NULL);
	engine_time = (unsigned long long)tv.tv_sec*1000000 + tv.tv_usec;
	engine_drawing_function();
	glXSwapBuffers(engine_p_display, engine_window);
	delta_tv.tv_sec  = engine_min_delta_tv.tv_sec;
	delta_tv.tv_usec = engine_min_delta_tv.tv_usec;

	single_key_up_flag = 0;
	xserver_fd = ConnectionNumber(engine_p_display);
	xserver_fd_p1 = xserver_fd + 1;

	event_loop:
		gettimeofday(&delta_tv, NULL);
		delta_tv.tv_sec  -= tv.tv_sec;
		delta_tv.tv_usec -= tv.tv_usec;
		if (delta_tv.tv_usec < 0)
		{
			delta_tv.tv_sec--;
			delta_tv.tv_usec += 1000000;
		}
		if ((delta_tv.tv_sec == engine_min_delta_tv.tv_sec &&
			delta_tv.tv_usec < engine_min_delta_tv.tv_usec) ||
			delta_tv.tv_sec  < engine_min_delta_tv.tv_sec)
		{
			delta_tv.tv_sec  = engine_min_delta_tv.tv_sec - delta_tv.tv_sec;
			delta_tv.tv_usec = engine_min_delta_tv.tv_usec - delta_tv.tv_usec;
			if (delta_tv.tv_usec < 0)
			{
				delta_tv.tv_usec += 1000000;
				delta_tv.tv_sec--;
			}
			FD_SET(xserver_fd, &fd_set_read);
			if
			(
				select(xserver_fd_p1, &fd_set_read, NULL, NULL, &delta_tv) > 0 &&
				XPending(engine_p_display)
			)
			{
				gettimeofday(&delta_tv, NULL);
				engine_time =(unsigned long long)(delta_tv.tv_sec)*1000000 +delta_tv.tv_usec;
				XNextEvent(engine_p_display, &event);
				switch (event.type)
				{
					case MotionNotify:
						DBGPRINTF
							(MOTION, "Pointer moved to %u, %u", event.xmotion.x, event.xmotion.y);
						engine_motion_handler(event.xmotion.x, event.xmotion.y);
					break;

					case KeyPress:
						keycode = event.xkey.keycode;
						if (keycode > 5 && keycode < engine_keys_names_number
							&& engine_keys_states[keycode] == engine_sc_key_up)
						{
							engine_keys_states[keycode] = 0;
							DBGPRINTF
								(KEY, "Key \"%s\" down", engine_keys_names[keycode]);
							engine_keys_handler(keycode, event.xkey.x, event.xkey.y, engine_sc_key_down);
						}
					break;

					case KeyRelease:
						keycode = event.xkey.keycode;
						if (5 < keycode && keycode < engine_keys_names_number
							&& engine_keys_states[keycode] == engine_sc_key_down)
							goto key_up;
					break;

					case ButtonPress:
						keycode = event.xbutton.button;
						if (0 < keycode && keycode < 6)
						{
							DBGPRINTF
								(KEY, "Button \"%s\" down", engine_keys_names[keycode]);
							engine_keys_handler
							(
								keycode,
								event.xbutton.x,
								event.xbutton.y,
								engine_sc_key_down
							);
						}
					break;

					case ButtonRelease:
						keycode = event.xbutton.button;
						if (0 < keycode && keycode < 6)
						{
							DBGPRINTF
								(KEY, "Button \"%s\" up", engine_keys_names[keycode]);
							engine_keys_handler
							(
								keycode,
								event.xbutton.x,
								event.xbutton.y,
								engine_sc_key_up
							);
						}
					break;

					case FocusIn:
						if (engine_focus_handler != NULL)
						{
							DBGPRINT(FOCUS, "Focus in");
							engine_focus_handler(engine_fs_focus_in);
						}
					break;

					case FocusOut:
						if (engine_focus_handler != NULL)
						{
							DBGPRINT(FOCUS, "Focus out");
							engine_focus_handler(engine_fs_focus_out);
						}
						single_key_up_flag = 0;
						for (keycode=0; keycode<engine_keys_names_number; keycode++)
							if (engine_keys_states[keycode] == engine_sc_key_down)
							{
								key_up:
								engine_keys_states[keycode] = 1;
								DBGPRINTF
									(KEY, "Key \"%s\" up", engine_keys_names[keycode]);
								engine_keys_handler(keycode, event.xkey.x, event.xkey.y, engine_sc_key_up);
								if (single_key_up_flag) goto event_loop;
							}
						single_key_up_flag = 1;
					break;

					case EnterNotify:
						if (engine_focus_handler != NULL)
						{
							DBGPRINT(FOCUS, "Pointer enter window");
							engine_focus_handler(engine_fs_pointer_enter);
						}
					break;

					case LeaveNotify:
						if (engine_focus_handler != NULL)
						{
							DBGPRINT(FOCUS, "Pointer leave window");
							engine_focus_handler(engine_fs_pointer_leave);
						}
					break;

					case ClientMessage:
						if (event.xclient.message_type == engine_atom_wm_protocols)
						{
							if (event.xclient.data.l[0]
								== engine_atoms[engine_wm_delete_window])
								goto event_loop_end;
							else
							if (event.xclient.data.l[0]
								== engine_atoms[engine_wm_save_yourself])
								goto event_loop_end;
							DBG(else)
								DBGPRINTF
								(
									WRNG,
									"Unhandled event, event.type=ClientMessage, \
event.xclient.message_type=atom_wm_protocols, event.xclient.data.l[0]=%li",
									event.xclient.data.l[0]
								);
						}
						DBG(else)
							DBGPRINTF
							(
								WRNG,
								"Unhandled event: event.type=ClientMessage, \
event.xclient.message_type=%li",
								event.xclient.message_type
							);
					DBG(break;
					default:)
						DBGPRINTF
						(
							WRNG,
							"Unhandled event, event.type=%i",
							event.type
						);
				}
				goto event_loop;
			}
		}
		gettimeofday(&tv, NULL);
		engine_time = (unsigned long long)tv.tv_sec*1000000 + tv.tv_usec;
		engine_drawing_function();
		glXSwapBuffers(engine_p_display, engine_window);
	goto event_loop;
	event_loop_end:
	return;
}

void engine_free()
{
	XCloseDisplay(engine_p_display);
}

char *engine_keys_names[] =
{
	"[0]", "MOUSE1", "MOUSE2", "MOUSE3", "MWHEELUP", "MWHEELDOWN", "[6]", "[7]", "[8]", "ESC",
	"1", "2", "3", "4", "5", "6", "7", "8", "9", "0",
	"-", "=", "BACKSPACE", "TAB", "q", "w", "e", "r", "t", "y",
	"u", "i", "o", "p", "[", "]", "ENTER", "LCTRL", "a", "s",
	"d", "f", "g", "h", "j", "k", "l", ";", "'", "`",

	"LSHIFT", "", "z", "x", "c", "v", "b", "n", "m", ",",
	".", "/", "RSHIFT", "NUM*", "LALT", "SPACE", "CAPS LOCK", "F1", "F2", "F3",
	"F4", "F5", "F6", "F7", "F8", "F9", "F10", "NUM LOCK", "SCROLL LOCK", "NUM7",
	"NUM8", "NUM9", "NUM-", "NUM4", "NUM5", "NUM6", "NUM+", "NUM1", "NUM2", "NUM3",
	"NUM0", "NUM.", "[92]", "[93]", "[94]", "F11", "F12", "HOME", "UP", "PAGE UP",

	"LEFT", "[101]", "RIGHT", "END", "DOWN", "PAGE DOWN", "INSERT", "DELETE", "NUMENTER", "RCTRL",
	"PAUSE", "PRINT SCREEN", "NUM/", "RALT", "[114]", "LWIN", "RWIN", "MENU", "[118]", "[119]"
};

unsigned engine_keys_names_number = sizeof(engine_keys_names)/sizeof(char *);
char engine_keys_states[sizeof(engine_keys_names)/sizeof(char *)];
